#define PROG_NAME "piketopnm"
#define PROG_DESC "convert AVC Pike F-100 RAW16 image files to PPM/PGM"
#define PROG_VERS "1.0"

#define piketopnm_C_COPYRIGHT "Copyright  2010 by the State University of Campinas (UNICAMP)"

/* Last edited on 2010-06-10 02:57:12 by stolfi */

#define PROG_HELP \
  PROG_NAME  " \\\n" \
  "  [ -debayer [ -interpolate ] ] \\\n" \
  "  [ -balance " frgb_parse_color_HELP " ] \\\n" \
  "  [ -verbose ] \\\n" \
  "  < {RAW16FILE} > {PPMFILE}"

#define PROG_INFO \
  "NAME\n" \
  "  " PROG_NAME " - " PROG_DESC "\n" \
  "\n" \
  "SYNOPSIS\n" \
  PROG_HELP "\n" \
  "\n" \
  "DESCRIPTION\n" \
  PROG_INFO_DESC "\n" \
  "\n" \
  "OPTIONS\n" \
  PROG_INFO_OPTS "\n" \
  "\n" \
  "DOCUMENTATION OPTIONS\n" \
  argparser_help_info_HELP_INFO "\n" \
  "\n" \
  "SEE ALSO\n" \
  "  pgm(5), ppm(5), pnm(5), pgmselect(1).\n" \
  "\n" \
  "AUTHOR\n" \
  "  Created in jun/2010 by Jorge Stolfi, IC-UNICAMP.\n" \
  "\n" \
  "  Info and experiments by Rafael Saracchini, IC-UNICAMP/MVL-UWE." \
  "\n" \
  "MODIFICATION HISTORY\n" \
  "  Option bla bla added by J. Stolfi, IC-UNICAMP.\n" \
  "\n" \
  "WARRANTY\n" \
  argparser_help_info_NO_WARRANTY "\n" \
  "\n" \
  "RIGHTS\n" \
  "  " piketopnm_C_COPYRIGHT ".\n" \
  "\n" \
  argparser_help_info_STANDARD_RIGHTS
  
#define PROG_INFO_DESC \
  "  The program reads a 1000x1000 raw image file produced by the AVC Pike F-100" \
  " camera, in {RAW16} mode.  It outputs a portable pixmap {PPMFILE}." \
  " It optionally applies a de-Bayering filter and other corrections."
  
#define PROG_INFO_OPTS \
  "  -debayer\n" \
  "    If this option is omitted, the output image will be" \
  " a PGM image with the R, G, and B samples interleaved in the 2x2" \
  " Bayer patern ((G,R),(B,G)).  If this option is present, it requests the unraveling of the" \
  " Bayer pattern.  The output will then be a three channel PPM" \
  " image, and each sample of the input file is moved to the appropriate" \
  " channel of the output.  The other two samples in the same pixel are set to" \
  " zero.\n" \
  "\n" \
  "  -interpolate\n" \
  "    This option is effective only if \"-debayer\" is also specified.  It" \
  " directs the program to interpolate any samples of the output image that" \
  " are missing in the input image.  It uses linear interpolation from the" \
  " nearest 2 or 4 samples of the same channel, depending on the channel" \
  " and on the pixel position.  If not specified, the missing samples" \
  " remain zero.\n" \
  "\n" \
  "  -balance " frgb_parse_color_HELP "\n" \
  "    This option specifies factors to be multiplied by the sample" \
  " values in the three RGB channels.  The factors are specified" \
  " as " frgb_parse_color_INFO "  This option is effective whether \"-debayer\" is used or not.  The default" \
  " is \"-balance 1 1 1\" (no color adjustment).\n" \
  "\n" \
  "  -verbose\n" \
  "    Produces diagnostic output."

#define _GNU_SOURCE
#include <stdlib.h>
#include <string.h>
#include <assert.h>

#include <bool.h>
#include <frgb_ops.h>
#include <affirm.h>
#include <argparser.h>
#include <jspnm_image.h>
#include <jspnm.h>

#include <jspnm_Pike_F100_image.h>

/* PROTOTYPES */

typedef struct options_t
  { /* Files: */
    char *input_file_name;   /* Input filename, or "-" i fnot given. */
    /* Filters: */
    bool_t debayer;      /* If true, perform Bayer channel separation. */
    bool_t interpolate;  /* If true, interpolate the missing samples in each channel. */
    frgb_t balance;      /* Color balance factors. */
    /* Diagnostics: */
    bool_t verbose;
  } options_t;
  /* Command line arguments. */

int main(int argc, char* argv[]);

options_t *piketopnm_parse_options(int argc, char **argv);
  /* Parses the command line arguments. */

/* REAL CODE */

int main(int argc, char** argv)
  {
    /* Parse command line arguments. */
    options_t *o = piketopnm_parse_options(argc, argv);

    /* Get input image: */
    pnm_image_t *img = pnm_Pike_F100_raw16_image_read(o->input_file_name, o->verbose);
    assert(img->chns == 1);
    
    if (o->debayer)
      { /* Separate pixels and apply debayering filter: */
        pnm_image_t *omg = pnm_Pike_F100_raw16_debayer(img, o->verbose);
        pnm_image_free(img);
        img = omg;
      }
    
    if (o->interpolate)
      { /* Interpolate missing samples: */
        if (img->chns != 3)
          { fprintf(stderr, "\"-interpolate\" ineffective without \"-debayer\""); }
        pnm_Pike_F100_raw16_interpolate_bayer(img, o->verbose);
      }
    
    frgb_t ones = frgb_Ones;
    if (! frgb_eq(&(o->balance), &ones))
      { /* Apply color balance factors: */
        pnm_Pike_F100_raw16_color_balance(img, o->balance.c, o->verbose);
      }
      
    /* Write output image: */
    bool_t forceplain = FALSE;
    pnm_image_fwrite(stdout, img, forceplain);
    
    /* All done: */
    if (o->verbose) { fprintf(stderr, "done.\n"); }
    exit(0);
  }

options_t *piketopnm_parse_options(int argc, char **argv)
  {
    options_t *o = (options_t *)notnull(malloc(sizeof(options_t)), "no mem");
    
    argparser_t *pp = argparser_new(stderr, argc, argv);
    argparser_set_help(pp, PROG_NAME " version " PROG_VERS ", usage:\n" PROG_HELP);
    argparser_set_info(pp, PROG_INFO);
    argparser_process_help_info_options(pp);

    o->debayer = argparser_keyword_present(pp, "-debayer");
    
    o->interpolate = argparser_keyword_present(pp, "-interpolate");
    
    if (argparser_keyword_present(pp, "-balance"))
      { o->balance = frgb_parse_color(pp); }
    else 
      { o->balance = frgb_Ones; }
    
    o->verbose = argparser_keyword_present(pp, "-verbose");
    
    /* Get optional input file name: */
    argparser_skip_parsed(pp);
    if (argparser_next(pp) != NULL) 
      { o->input_file_name = argparser_get_next(pp); }
    else
      { o->input_file_name = "-"; }

    /* Check for extraneous arguments: */
    argparser_finish(pp);
    
    return o;
  }
