#define PROG_NAME "test_align"
#define PROG_DESC "test of {float_image_align.h}"
#define PROG_VERS "1.0"

/* Last edited on 2009-07-09 22:46:41 by stolfi */ 
/* Created on 2007-07-11 by J. Stolfi, UNICAMP */

#define test_align_COPYRIGHT \
  "Copyright  2007  by the State University of Campinas (UNICAMP)"

#define _GNU_SOURCE
#include <stdio.h>
#include <stdlib.h>
#include <stdint.h>
#include <string.h>
#include <math.h>

#include <float_image.h>
#include <float_pnm_image.h>
#include <float_image_align.h>
#include <float_image_mismatch.h>
#include <float_image_interpolate.h>
#include <float_pnm_image_io.h>
#include <wt_table.h>
#include <indexing.h>
#include <r2.h>
#include <r2x2.h>
#include <bool.h>
#include <jsfile.h>
#include <jsrandom.h>
#include <affirm.h>
#include <assert.h>

int main(int argn, char **argv);

void test_align_one(char *ext, char *fname, bool_t bias, int NI, int chn, r2_t cmp_rad, r2_t adj_rad);
  /* Tests alignment for {NI} images with extension {ext},
    mismatch function {fname}, channel {chn}, window radius {cmp_rad},
    search radius {adj_rad}. */

float_image_t *test_align_read_image(char *prefix, int k, char *ext);
  /* Reads a PBM/PGM/PPM image file with name "{prefix}-{KK}.{EXT}", 
    where {KK} is the two-digit value of {k}, and converts it to a float
    image with samples in the range [0_1].. */

void test_align_write_image(char *prefix, int k, float_image_t *img);
  /* Writes {img} to a PGM/PPM image file with name "{prefix}-{KK}.{EXT}", 
    where {KK} is the two-digit value of {k}, ad {ext} is 
    "pgm" or "ppm" depending on the channel count of {img}. */

void test_align_choose_optimum(int NI, r2_t pini[], r2_t adj_rad, r2_t popt[]);
  /* Stores into {popt[0..NI-1]} the optimum solution for the test
    with {NI} images, initial guess {pini[0..NI-1]}, and search radius {adj_rad}.
    makes sure that {popt-pini} is balanced. */

bool_t test_align_ckeck_result(int NI, r2_t psol[], r2_t pini[], r2_t popt[], r2_t adj_rad); 
  /* Check correctness of solution {psol} against known optimum {popt} (if not NULL).
    The solution is correct if every point {psol[i]} differs from {popt[i]}
    by the same displacement vector. Also checks that the displacements from {pini[i]}
    to {psol[i]} add to {(0,0)}. */

float_image_t *test_align_make_displaced_image(float_image_t *img, r2_t q);
  /* Returns a copy {cpy} of {img}, displaced so that the center of {img} 
    corresponds to point {q} of {cpy}. */

void test_image_debug_points(char *label, int NI, r2_t o[], r2_t p[], double f2, double d2);
  /* Prints the points {p[0..NI-1]} and the corresponding goal function value {Q}. 
    If {o} is not NULL, prints also the difference between each {p[i]} and the 
    corresponding {o[i]}. */

void test_image_plot_goal
  ( float_image_align_mismatch_t *f2, 
    int NI, 
    int scale, 
    int hwx, 
    double wx[], 
    int hwy, 
    double wy[], 
    r2_t p[],
    r2_t *adj_rad
  );
  /* Writes a  file "out/f2.dat" with a random 2D slice of the goal function
    {f2(NI, scale, hwx, wx, hwy, wy, q)}, where {q} ranges in the neighborhood
    of {p}, with each {q[i]} inside the rectangle defined by {adj_rad}. */

int main(int argc, char **argv)
  {
    srandom(4615*417);
    
    /* Parse arguments: */
    char *rest;

    int na;
    na = 1;
    char *ext = argv[na++];                    /* Image extension. */
    char *fname = argv[na++];                  /* Function name. */
    int NI = strtol(argv[na++],&rest,10);      /* Number of images to align. */
    demand((*rest) == 0, "invalid NI");
    bool_t bias = strtol(argv[na++],&rest,10); /* 1 to add squared distance bias. */
    demand((*rest) == 0, "invalid bias");
    r2_t cmp_rad;
    cmp_rad.c[0] = strtod(argv[na++],&rest);   /* Radius of comparison window. */
    cmp_rad.c[1] = strtod(argv[na++],&rest);   /* Radius of comparison window. */
    demand((*rest) == 0, "invalid window radius");
    r2_t adj_rad;
    adj_rad.c[0] = strtod(argv[na++],&rest);  /* Search radius. */
    adj_rad.c[1] = strtod(argv[na++],&rest);  /* Search radius. */
    demand((*rest) == 0, "invalid search radius");

    demand(argc == na, "wrong argument count");
    
    test_align_one(ext, fname, bias, NI, 0, cmp_rad, adj_rad);
    return 0;
  }
  
void test_align_one(char *ext, char *fname, bool_t bias, int NI, int chn, r2_t cmp_rad, r2_t adj_rad)
  { 
    fprintf(stderr, "testing with channel %d of %d %s synthetic images...\n", chn, NI, ext);
    fprintf(stderr, "function = %s  bias = %c\n", fname, "FT"[bias]);
    
    bool_t debug_points = FALSE; /* TRUE to print every probe point. */
    
    r2_t popt[NI]; /* Actual optimum alignment. */
    r2_t pini[NI]; /* Initial alignment. */

    /* Create the window weight tables: */
    int hwx = (int)ceil(fabs(cmp_rad.c[0])), nwx = 2*hwx + 1;
    double wx[nwx];
    wt_table_fill_hann(nwx, wx);
    
    int hwy = (int)ceil(fabs(cmp_rad.c[1])), nwy = 2*hwy + 1;
    double wy[nwy];
    wt_table_fill_hann(nwy, wy);

    int NC, NX, NY;         /* Image dimensions (actual or nominal). */
    float_image_t *img[NI]; /* Images to be aligned (or NULL). */
    bool_t cmp_opt;         /* Should the solution be compared to {popt}? */
    
    void img_eval(int i, r2_t *p, int hw, double w[], double v[]);
      /* A goal function that evaluates one sample of image number {i}
        at the point {p+(ix,iy)}, with suitable interpolation. */
    
    auto double f2_indiff(int NI, int scale, int fhwx, double fwx[], int fhwy, double fwy[], r2_t p[]);
      /* A goal function that returns 1.0 always.  If {bias} is true,
        the minimum should be at {pini}. */
    
    auto double f2_images(int NI, int scale, int fhwx, double fwx[], int fhwy, double fwy[], r2_t p[]);
      /* A goal function that compares actual images {img[0..NI-1]},
        displaced so that the points {p[0..NI-1]} are aligned. If
        {bias} is FALSE, the minimum should be near {popt}. */
    
    auto double f2_points(int NI, int scale, int fhwx, double fwx[], int fhwy, double fwy[], r2_t p[]);
      /* A goal function that ignores the images and simply returns
        the sum of the squared differences from {p[0..NI-1]} to the
        optimum {popt[0..NI-1]}. If {bias} is FALSE, the
        minimum should be very close to {popt}. */
    
    int i;
       
    /* Choose the goal function {f2} accordng to {fname}. */
    /* Also set auxiliary parameters {NC,NX,NY,img[],pini[],popt[]}: */
    float_image_align_mismatch_t *f2 = NULL;
    if (strcmp(fname, "indiff") == 0)
      { f2 = &f2_indiff;
        NC = (strcmp(ext, "ppm") == 0 ? 3 : 1);
        NX = 32;
        NY = 32;
        /* Choose the initial guess {pini}: */
        for (i = 0; i < NI; i++) { pini[i] = (r2_t){{ 0.5*NX, 0.5*NY }}; }
        /* For this function, the optimum {popt} is the initial point {pini}: */
        for (i = 0; i < NI; i++) { popt[i] = pini[i]; }
        cmp_opt = bias; /* If {bias} is true, the solution should be {popt}. */
        /* This function does not use any images: */
        for (i = 0; i < NI; i++) { img[i] = NULL; }
      }
    else if (strcmp(fname, "points") == 0)
      { f2 = &f2_points;
        NC = (strcmp(ext, "ppm") == 0 ? 3 : 1);
        NX = 32;
        NY = 32;
        /* Choose the initial guess {pini}: */
        for (i = 0; i < NI; i++) { pini[i] = (r2_t){{ 0.5*NX, 0.5*NY }}; }
        /* Choose an optimum {popt} near the initial point {pini}: */
        test_align_choose_optimum(NI, pini, adj_rad, popt);
        cmp_opt = ! bias; /* If {bias} is false, the solution should be {popt}. */
        /* This function does not use any images: */
        for (i = 0; i < NI; i++) { img[i] = NULL; }
      }
    else if (strcmp(fname, "images") == 0)
      { f2 = &f2_images;
        /* Read the basic image. */
        float_image_t *img0 = test_align_read_image("in/img", 0, ext);
        float_image_get_size(img0, &NC, &NX, &NY);
        /* Choose the initial guess {pini}: */
        for (i = 0; i < NI; i++) { pini[i] = (r2_t){{ 0.5*NX, 0.5*NY }}; }
        /* Choose the optimum displacements {popt} near the initial point {pini}: */
        test_align_choose_optimum(NI, pini, adj_rad, popt);
        cmp_opt = ! bias; /* If {bias} is false, the solution should be {popt}. */
        /* Generate the displaced images: */
        for (i = 0; i < NI; i++) 
          { img[i] = test_align_make_displaced_image(img0, popt[i]);
            test_align_write_image("out/img", i, img[i]);
          }
        float_image_free(img0);
      }
    else 
      { demand(FALSE, "invalid function name"); }
    
    double f2opt = f2(NI, 0, hwx, wx, hwy, wy, popt);
    double d2opt = float_image_align_rel_disp_sqr(NI, popt, pini, &adj_rad);
    test_image_debug_points("actual optimum", NI, NULL, popt, f2opt, d2opt);
    
    double f2ini = f2(NI, 0, hwx, wx, hwy, wy, pini);
    double d2ini = float_image_align_rel_disp_sqr(NI, pini, pini, &adj_rad);
    test_image_debug_points("initial guess",  NI, popt, pini, f2ini, d2ini);
    
    /* Plot the goal function in the neighborhood of the initial guess: */ 
    test_image_plot_goal(f2, NI, 0, hwx, wx, hwy, wy, pini, &adj_rad);
    
    /* Call the optimizer: */
    r2_t psol[NI];  /* Corresponding points in each image. */
    for (i = 0; i < NI; i++) { psol[i] = pini[i]; }
    float_image_align_single_scale(NI, 0, f2, hwx, wx, hwy, wy, psol, adj_rad);
    
    double f2sol = f2(NI, 0, hwx, wx, hwy, wy, psol);
    double d2sol = float_image_align_rel_disp_sqr(NI, psol, pini, &adj_rad);
    test_image_debug_points("computed optimum", NI, popt, psol, f2sol, d2sol);
    
    bool_t ok = test_align_ckeck_result(NI, psol, pini, (cmp_opt ? popt : NULL), adj_rad);

    if(! ok) { exit(1); }
    
    return;
    
    double f2_indiff(int NI, int scale, int fhwx, double fwx[], int fhwy, double fwy[], r2_t p[])
      { demand(NI == NI, "duh?");
        demand(scale == 0, "dah?");
        demand((fhwx == hwx) && (fhwy == hwy), "deh?");
        demand((fwx == wx) && (fwy == wy), "doh?");
        
        double f2_val = 1.0;
        if (debug_points) 
          { double d2 = float_image_align_rel_disp_sqr(NI, popt, p, &adj_rad);
            test_image_debug_points("probe point   ", NI, popt, p, f2_val, d2);
          }
        return f2_val;
      }
    
    double f2_points(int NI, int scale, int fhwx, double fwx[], int fhwy, double fwy[], r2_t p[])
      { demand(NI == NI, "duh?");
        demand(scale == 0, "dah?");
        demand((fhwx == hwx) && (fhwy == hwy), "deh?");
        demand((fwx == wx) && (fwy == wy), "doh?");
        /* Compute sum of square distances from actual optimum {opt}: */
        double f2_val = 0;
        int i;
        for (i = 0; i < NI; i++) 
          { r2_t ok = popt[i];
            r2_t d;
            r2_sub(&(p[i]), &ok, &d);
            f2_val += r2_norm_sqr(&d);
          }
        /* Scale {f2_val} so that it dominates the displacement term: */
        f2_val *= 1.0e7;
        if (debug_points) 
          { double d2 = float_image_align_rel_disp_sqr(NI, popt, p, &adj_rad);
            test_image_debug_points("probe point   ", NI, popt, p, f2_val, d2);
          }
        return f2_val;
      }
      
    double f2_images(int NI, int scale, int fhwx, double fwx[], int fhwy, double fwy[], r2_t p[])
      { demand(NI == NI, "duh?");
        demand(scale == 0, "dah?");
        demand((fhwx == hwx) && (fhwy == hwy), "deh?");
        demand((fwx == wx) && (fwy == wy), "doh?");
        
        double f2_val = float_image_mismatch_var(NI, &img_eval, hwx, wx, hwy, wy, p);
        f2_val *= 1.0e-7;
        if (debug_points) 
          { double d2 = float_image_align_rel_disp_sqr(NI, popt, p, &adj_rad);
            test_image_debug_points("probe point   ", NI, popt, p, f2_val, d2);
          }
        return f2_val;
      }
    
    void img_eval(int i, r2_t *p, int hw, double w[], double v[])
     {
       int ix, iy;
       int k = 0;
       int order = 1;
       ix_reduction_t red = ix_reduction_MIRROR;
       for (iy = -hwy; iy <= +hwy; iy++)
         { for (ix = -hwx; ix <= +hwx; ix++)
             { v[k] = float_image_interpolate_sample(img[i], chn, p->c[0] + ix, p->c[1] + iy, order, red);
               w[k] = 1.0;
               k++;
             }
         }
     }
  }
    
void test_align_choose_optimum(int NI, r2_t pini[], r2_t adj_rad, r2_t popt[])
  {
    if (NI == 0) { return; }
    
    int a;
    for (a = 0; a < 2; a++)
      { double ra = 0.50*adj_rad.c[a]; /* Search radius along axis {a}. */
        double da[NI];            /* Displacements along axis {a}. */
        /* Set {da[0..NI-1]} to random displcements: */
        int sumd = 0; /* Sum of all {da[0..NI-1]}. */
        int i;
        for (i = 0; i < NI; i++) { da[i] = (2*drandom()-1)*ra; sumd += da[i]; }
        /* Adjust {da} so that it is balanced: */
        int corr = -sumd/NI;
        for (i = 0; i < NI; i++) { da[i] += corr; }
        /* Convert {da[0..NI-1]} to displacements {iopt[0..NI-1].c[a]} from {p}: */
        for (i = 0; i < NI; i++) { popt[i].c[a] = pini[i].c[a] + da[i]; }
      }
  }
    
void test_image_plot_goal
  ( float_image_align_mismatch_t *f2, 
    int NI, 
    int scale, 
    int hwx, 
    double wx[], 
    int hwy, 
    double wy[], 
    r2_t p[],
    r2_t *adj_rad
  )
  {
    /* Choose unit vectors {u[0..NI-1]} in {R^2}, with sum {(0,0)}: */
    int i;
    r2_t u[NI];
    /* Throw a bunch of random unit vectors: */
    r2_t usum = (r2_t){{ 0, 0 }};
    for (i = 0; i < NI; i++) { r2_throw_dir(&(u[i])); r2_add(&(u[i]), &usum, &usum); }
    while (r2_norm(&usum) > 1.0e-5)
      { /* Shift {u[0..NI-1]} so that it has zero sum, and normalize: */
        r2_t uavg;
        r2_scale(1.0/NI, &usum, &uavg);
        usum = (r2_t){{ 0, 0 }};
        for (i = 0; i < NI; i++)
          { r2_sub(&(u[i]), &uavg, &(u[i]));
            double u2 = r2_dir(&(u[i]), &(u[i]));
            if (u2 < 1.0e-4) { r2_throw_dir(&(u[i])); }
            r2_add(&(u[i]), &usum, &usum);
          }
      }
      
    /* Choose unit vectors {v[i]} in {R^2}, with sum {(0,0)}, perp to {u[i]}: */
    r2_t v[NI];
    for (i = 0; i < NI; i++) { r2_cross(&(u[i]), &(v[i])); }
    
    /* Sweep the {p,u,v} plane and plot: */
    char *fname = "out/f2.dat";
    FILE *fpl = open_write(fname, TRUE);
    
    int ns = 20; /* Number of steps in each direction, in each sense. */
    int iu, iv;
    r2_t q[NI]; /* Probe points. */
    for (iu = -ns; iu <= +ns; iu++)
      { double du = ((double)iu)/((double)ns);
        for (iv = -ns; iv <= +ns; iv++)
          { double dv = ((double)iv)/((double)ns);
            /* Compute the probe points {q[0..NI-1]}. */
            for(i = 0; i < NI; i++)
              { r2_mix(du,&(u[i]), dv, &(v[i]), &(q[i])); 
                r2_add(&(p[i]), &(q[i]), &(q[i])); 
              }
            /* Evaluate the function and plot: */
            double f2_val = f2(NI, scale, hwx, wx, hwy, wy, q);
            fprintf(fpl, "%+9.6f %+9.6f  %12.6f\n", du, dv, f2_val); 
          }
        /* Blank line between scanlines, for {gnuplot}: */
        fprintf(fpl, "\n"); 
      }

    fclose(fpl);
  }

bool_t test_align_ckeck_result(int NI, r2_t psol[], r2_t pini[], r2_t popt[], r2_t adj_rad)
  {
    int ind = 0; /* Indentation */
    fprintf(stderr, "%*schecking the solution\n", ind, "");
    r2_t dopt; /* Common displacement between solution and optimum points. */
    r2_t dini; /* Total displacement between solution and initial points. */
    bool_t valid = TRUE;
    int i, a;
    for (i = 0; i < NI; i++)
      { r2_t sopt, sini;
        r2_sub(&(psol[i]), &(popt[i]), &sopt);
        r2_sub(&(psol[i]), &(pini[i]), &sini);
        
        fprintf(stderr, "%*s  %02d ( %9.4f %9.4f )", ind, "", i, psol[i].c[0], psol[i].c[1]);
        fprintf(stderr, " = ini + ( %+9.4f %+9.4f )", sini.c[0], sini.c[1]);
        fprintf(stderr, " = opt + ( %+9.4f %+9.4f )", sopt.c[0], sopt.c[1]);
        if (i == 0)
          { /* Save displacement from optimum: */
            dopt = sopt;
            dini = sini;
          }
        else
          { /* Check whether displacements from optimum are all equal: */
            for (a = 0; a < 2; a++)
              { if (fabs(sopt.c[a] - dopt.c[a]) > 1.0e-4*adj_rad.c[a])
                  { fprintf(stderr, " ** err[%d] = %+9.4f", a, sopt.c[a] - dopt.c[a]);
                    valid = FALSE;
                  }
              }
            /* Accumulate displacement from initial point: */
            r2_add(&sini, &dini, &dini);
          }
        fprintf(stderr, "\n");
      }
    /* Check whether displacement from initial point add to (0,0): */
    fprintf(stderr, "%*s  total %28s ( %+9.4f %+9.4f )", ind, "", "", dini.c[0], dini.c[1]);
    for (a = 0; a < 2; a++)
      { if (fabs(dini.c[a]) > 1.0e-4*adj_rad.c[a])
          { fprintf(stderr, " ** err[%d] = %+9.4f", a, dini.c[a]);
            valid = FALSE;
          }
      }
    fprintf(stderr, "\n");
    return valid;
  }

float_image_t *test_align_read_image(char *prefix, int k, char *ext)
  { char *fname = NULL;
    asprintf(&fname, "%s-%02d.%s", prefix, k, ext);
    pnm_image_t *pim = pnm_image_read(fname, TRUE);
    float_image_t *fim = float_image_from_pnm_image(pim, NULL, NULL, TRUE, TRUE); 
    pnm_image_free(pim);
    free(fname);
    return fim;
  }

void test_image_debug_points(char *label, int NI, r2_t o[], r2_t p[], double f2, double d2)
  { 
    int ind = 4; /* Indentation */
    fprintf(stderr, "%*s%s\n", ind, "", label);
    int i;
    for (i = 0; i < NI; i++)
      { fprintf(stderr, "%*s  %02d ( %9.4f %9.4f )", ind, "", i, p[i].c[0], p[i].c[1]);
        if (o != NULL)
          { fprintf(stderr, " = opt + ( %+9.4f  %+9.4f )", p[i].c[0]-o[i].c[0], p[i].c[1]-o[i].c[1]); }
        fprintf(stderr, "\n");
      }
    fprintf(stderr, "%*s  f2 = %22.10f  d2 = %22.10f  Q2 = %22.10f\n", ind, "", f2, d2, f2+d2);
    fprintf(stderr, "\n");
  }

float_image_t *test_align_make_displaced_image(float_image_t *img, r2_t q)
  {
    /* Get image dimensions: */
    int NC = img->sz[0];
    int NX = img->sz[1];
    int NY = img->sz[2];
    
    /* Compute the  displacement vector {d}: */
    r2_t o = (r2_t){{ 0.5*NX, 0.5*NY }};
    r2_t d; r2_sub(&q, &o, &d);
    
    /* Get an image with the right size: */
    float_image_t *cpy = float_image_new(NC, NX, NY);
    
    /* Allocate a pixel's worth of samples: */
    double v[NC];
    float f[NC];
    
    /* Scan pixels: */
    int ix, iy;
    int order = 1;
    ix_reduction_t red = ix_reduction_PXMIRR;
    for (iy = 0; iy < NY; iy++)
      { for (ix = 0; ix < NX; ix++)
          { /* Get coordinates of pixel center, relative to image center: */
            r2_t p = (r2_t){{ ix + 0.5, iy + 0.5 }};
            /* Get coordinates of source point: */
            r2_t q; r2_sub(&p, &d, &q);
            /* Interpolate image: */
            float_image_interpolate_pixel(img, q.c[0], q.c[1], order, red, v);
            /* Store into pixel: */
            int i; for (i = 0; i < NC; i++) { f[i] = v[i]; }
            float_image_set_pixel(cpy, ix, iy, f);
          }
      }
    return cpy;
  }

void test_align_write_image(char *prefix, int k, float_image_t *img)
  {
    char *ext = (img->sz[0] == 3 ? "ppm" : "pgm");
    
    char *fname = NULL;
    asprintf(&fname, "%s-%02d.%s", prefix, k, ext);
    float_pnm_image_write(fname, img, 0.4500, 0.0327, TRUE, TRUE, TRUE);
    free(fname);
  }
